@@ -746,10 +746,10 @@ static int the_end (reiserfs_filsys_t * fs)
 
     mark_filesystem_consistent (fs);
     clear_buffer_do_not_flush (fs->fs_super_bh);
-    
-    if (misc_device_mounted(fs->fs_file_name, MF_RO)) {
-	reiserfs_warning(stderr, "\nThe partition is mounted ro. It is better "
-			 "to umount and mount it again.\n\n");
+
+    if (fsck_data(fs)->mounted == MF_RO) {
+	reiserfs_warning(stderr, "\nThe partition is mounted ro. It "
+			 "is better to umount and mount it again.\n\n");
 	ret = EXIT_REBOOT;
     }
 
@@ -764,18 +764,20 @@ static int the_end (reiserfs_filsys_t * fs)
 }
 
 /* check umounted or read-only mounted filesystems only */
-void prepare_fs_for_check(reiserfs_filsys_t * fs) {
+static void prepare_fs_for_check(reiserfs_filsys_t * fs) {
     /* The method could be called from auto_check already. */
     if (fs->fs_flags == O_RDWR) 
 	return;
 
     reiserfs_reopen (fs, O_RDWR);
     
-    if (misc_device_mounted(fs->fs_file_name, 0)) {
-	if (!misc_device_mounted(fs->fs_file_name, MF_RO)) {
+    fsck_data(fs)->mounted = misc_device_mounted(fs->fs_file_name);
+    
+    if (fsck_data(fs)->mounted > 0) {
+	if (fsck_data(fs)->mounted == MF_RW) {
 	    fsck_progress ("Partition %s is mounted with write permissions, "
 		"cannot check it\n", fs->fs_file_name);
-	    reiserfs_close (fs);
+	    reiserfs_close(fs);
 	    exit(EXIT_USER);
 	}
 	
@@ -1093,7 +1095,7 @@ static void check_fs (reiserfs_filsys_t * fs)
 	stage_report (5, fs);
 
 	if (fsck_mode(fs) != FSCK_CHECK) {
-		if (misc_device_mounted(fs->fs_file_name, MF_RO)) {
+		if (misc_device_mounted(fs->fs_file_name) == MF_RO) {
 			reiserfs_warning(stderr, "\nThe partition is mounted ro. It is better "
 					 "to umount and mount it again.\n\n");
 			retval = EXIT_REBOOT;
@@ -1188,9 +1190,10 @@ int main (int argc, char * argv [])
 {
     char * file_name;
     struct fsck_data * data;
-    struct rlimit rlim = {0xffffffff, 0xffffffff};
-    int retval;
+    struct rlimit rlim = {RLIM_INFINITY, RLIM_INFINITY};
     char *width;
+    int retval;
+    int errno;
     
     width = getenv("COLUMNS");
     if ( width )
@@ -1254,17 +1257,22 @@ int main (int argc, char * argv [])
     	if (open_devices_for_rollback (file_name, data) == -1)
     	    exit(EXIT_OPER);
     } else {
-	fs = reiserfs_open (file_name, O_RDONLY, 0, data, 
+	fs = reiserfs_open (file_name, O_RDONLY, &errno, data, 
 	    data->mode != FSCK_SB);
 
 	if (data->mode != FSCK_SB) {
-	    if (no_reiserfs_found (fs)) 
-    	    	die ("Failed to open the filesystem.\n\n"
-		    "If the partition table has not been changed, and the partition is\n"
-		    "valid  and  it really  contains  a reiserfs  partition,  then the\n"
-		    "superblock  is corrupted and you need to run this utility with\n"
-		    "--rebuild-sb.\n");
-
+	    if (no_reiserfs_found (fs)) {
+		if (errno) {
+		    die ("Failed to open the device '%s': %s\n\n", 
+			 file_name, strerror(errno));
+		} else {
+		    die ("Failed to open the filesystem.\n\n"
+			 "If the partition table has not been changed, and the partition is\n"
+			 "valid  and  it really  contains  a reiserfs  partition,  then the\n"
+			 "superblock  is corrupted and you need to run this utility with\n"
+			 "--rebuild-sb.\n");
+		}
+	    }
 	    if (fsck_skip_journal (fs) && 
 		!is_reiserfs_jr_magic_string (fs->fs_ondisk_sb)) 
 	    {
